/*GPL*START*
 * 
 * Copyright (C) 1998 by Johannes Overmann <overmann@iname.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * *GPL*END*/  

#include <qtooltip.h>
#include <qlabel.h>
#include <qmsgbox.h>
#include <qtabbar.h>
#include <qpushbt.h>
#include <stdlib.h>
#include <qgrpbox.h>
#include <qlayout.h>
#include <ctype.h>
#include <qbttngrp.h>
#include <qpixmap.h>

#include "mainwidget.h"
#include "tappconfig.h"
#include "loadgraph.h"

#include "qload.xpm"

// sort slots:

void MainWidget::sort_load1() {
   if(m.num()>1) {
      if(sort==S_LOAD1) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_load1_r);
	 sort=S_LOAD1_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_load1);
	 sort=S_LOAD1;
      }
      emit mach_changed();
   }
}


void MainWidget::sort_load2() {
   if(m.num()>1) {
      if(sort==S_LOAD2) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_load2_r);
	 sort=S_LOAD2_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_load2);
	 sort=S_LOAD2;
      }
      emit mach_changed();
   }
}


void MainWidget::sort_load3() {
   if(m.num()>1) {
      if(sort==S_LOAD3) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_load3_r);
	 sort=S_LOAD3_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_load3);
	 sort=S_LOAD3;
      }
      emit mach_changed();
   }
}


void MainWidget::sort_name() {
   if(m.num()>1) {
      if(sort==S_NAME) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_name_r);
	 sort=S_NAME_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_name);
	 sort=S_NAME;
      }
      emit mach_changed();
   }
}


void MainWidget::sort_ram() {
   if(m.num()>1) {
      if(sort==S_RAM) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_ram_r);
	 sort=S_RAM_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_ram);
	 sort=S_RAM;
      }
      emit mach_changed();
   }
}


void MainWidget::sort_group() {
   if(m.num()>1) {
      if(sort==S_GROUP) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_group_r);
	 sort=S_GROUP_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_group);
	 sort=S_GROUP;
      }
      emit mach_changed();
   }
}


void MainWidget::sort_arch() {
   if(m.num()>1) {
      if(sort==S_ARCH) {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_arch_r);
	 sort=S_ARCH_R;
      } else {
	 isort(m.data(), m.num(), sizeof(Machine), machcmp_arch);
	 sort=S_ARCH;
      }
      emit mach_changed();
   }
}


void MainWidget::resort() {
   switch(sort) {
    case S_RANDOM:
    case S_LOAD1:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_load1);
      break;
    case S_LOAD1_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_load1_r);
      break;
    case S_LOAD2:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_load2);
      break;
    case S_LOAD2_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_load2_r);
      break;
    case S_LOAD3:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_load3);
      break;
    case S_LOAD3_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_load3_r);
      break;
    case S_NAME:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_name);
      break;
    case S_NAME_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_name_r);
      break;
    case S_RAM:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_ram);
      break;
    case S_RAM_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_ram_r);
      break;
    case S_GROUP:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_group);
      break;
    case S_GROUP_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_group_r);
      break;
    case S_ARCH:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_arch);
      break;
    case S_ARCH_R:
      isort(m.data(), m.num(), sizeof(Machine), machcmp_arch_r);
      break;
   }
   emit mach_changed();
}



// setup:

void MainWidget::createSetup() {
   int le_height = 24; 
   setup = new QWidget();
   QBoxLayout *lay0 = new QBoxLayout(setup, QBoxLayout::Down, 5);
   QFrame *fra = new QFrame(setup);
   fra->setLineWidth(1);
   fra->setFrameStyle(QFrame::Panel|QFrame::Raised);
   lay0->addWidget(fra, 1);
   QBoxLayout *lay1 = new QBoxLayout(QBoxLayout::LeftToRight, 5);
   lay0->addLayout(lay1, 0);
   QPushButton *b_ok = new QPushButton("OK", setup);
   QPushButton *b_apply = new QPushButton("Apply", setup);
   QPushButton *b_default = new QPushButton("Default", setup);
   QPushButton *b_reset = new QPushButton("Reset", setup);
   QPushButton *b_cancel = new QPushButton("Cancel", setup);
   QPushButton *b_save = new QPushButton("Save", setup);
   b_autosave = new QCheckBox("Auto save on exit", setup);
   lay1->addWidget(b_ok,10);
   lay1->addWidget(b_apply,10);
   lay1->addStretch(5);
   lay1->addWidget(b_reset, 10);
   lay1->addWidget(b_default, 10);
   lay1->addWidget(b_save,10);
   lay1->addWidget(b_autosave, 27);
   lay1->addStretch(3);
   lay1->addWidget(b_cancel,10);
   lay1->addStrut(b_ok->sizeHint().height());
   connect(b_apply, SIGNAL(clicked()), SLOT(applyFromSetup()));
   connect(b_ok, SIGNAL(clicked()), SLOT(applyFromSetup()));
   connect(b_ok, SIGNAL(clicked()), SLOT(toggle_setup()));
   connect(b_cancel, SIGNAL(clicked()), SLOT(toggle_setup()));
   connect(b_cancel, SIGNAL(clicked()), SLOT(resetSetup()));
   connect(b_default, SIGNAL(clicked()), SLOT(defaultSetup()));
   connect(b_save, SIGNAL(clicked()), SLOT(applyAndSaveItems()));
   connect(b_reset, SIGNAL(clicked()), SLOT(resetSetup()));
   setup->setCaption(tApp->getString("application-name") + " Setup");
   setup->setMinimumSize(576,340);
   setup->resize(minimumSize());
   QToolTip::add(b_default, "Restore factory default settings");
   QToolTip::add(b_reset, "Restore last saved settings");
   QToolTip::add(b_save, "Save and apply current settings");   
   
   QBoxLayout *lay2 = new QBoxLayout(fra, QBoxLayout::Down, 5);
   QLabel *l_rup = new QLabel("Ruptime", fra);
   QLabel *l_grp = new QLabel("Group File", fra);
   QLabel *l_mem = new QLabel("Memory File", fra);
   QLabel *l_pop = new QLabel("Popup File", fra);
   QLabel *l_arc = new QLabel("Architecture file", fra);
   l_rup->setMinimumSize(l_rup->sizeHint());
   l_grp->setMinimumSize(l_grp->sizeHint());
   l_mem->setMinimumSize(l_mem->sizeHint());
   l_pop->setMinimumSize(l_pop->sizeHint());
   l_arc->setMinimumSize(l_arc->sizeHint());
   QGridLayout *lay3 = new QGridLayout(5, 2, 5);
   lay2->addLayout(lay3, 1);
   lay3->addWidget(l_rup, 0, 0);
   lay3->addWidget(l_grp, 1, 0);
   lay3->addWidget(l_mem, 2, 0);
   lay3->addWidget(l_pop, 3, 0);
   lay3->addWidget(l_arc, 4, 0);
   lay3->setColStretch(0,0);
   lay3->setColStretch(1,1);
   le_ruptime = new QLineEdit(fra);
   le_groups = new QLineEdit(fra);
   le_memory = new QLineEdit(fra);
   le_pop = new QLineEdit(fra);
   le_arc = new QLineEdit(fra);
   le_ruptime->setMinimumSize(QSize(100, le_height));
   le_ruptime->setMaximumSize(QSize(10000, le_height));
   le_groups->setMinimumSize(QSize(100, le_height));
   le_groups->setMaximumSize(QSize(10000, le_height));
   le_memory->setMinimumSize(QSize(100, le_height));
   le_memory->setMaximumSize(QSize(10000, le_height));
   le_pop->setMinimumSize(QSize(100, le_height));
   le_pop->setMaximumSize(QSize(10000, le_height));
   le_arc->setMinimumSize(QSize(100, le_height));
   le_arc->setMaximumSize(QSize(10000, le_height));
   lay3->addWidget(le_ruptime, 0, 1);
   lay3->addWidget(le_groups, 1, 1);
   lay3->addWidget(le_memory, 2, 1);
   lay3->addWidget(le_pop   , 3, 1);
   lay3->addWidget(le_arc   , 4, 1);
   
   le_update = new QLineEdit(fra);
   le_update->setMinimumSize(QSize(50, le_height));
   le_update->setMaximumSize(QSize(10000, le_height));
   QLabel *l_update = new QLabel("Update interval", fra); 
   QLabel *l_sec = new QLabel("sec", fra); 
   l_update->setMinimumSize(l_update->sizeHint());
   l_sec->setMinimumSize(l_sec->sizeHint());
   
   
   // setup widgets
   
   QGroupBox *graph_box = new QGroupBox("Graph", fra);
   QButtonGroup *graph_bg = new QButtonGroup;
   QLabel *rb_graph_lab = new QLabel("One graph for:", graph_box);
   rb_graph_sep = new QRadioButton("each Host", graph_box);
   rb_graph_grp = new QRadioButton("each Group", graph_box);
   rb_graph_one = new QRadioButton("all Hosts", graph_box);
   rb_graph_lab->setMinimumSize(rb_graph_lab->sizeHint());
   rb_graph_sep->setMinimumSize(rb_graph_sep->sizeHint());
   rb_graph_grp->setMinimumSize(rb_graph_grp->sizeHint());
   rb_graph_one->setMinimumSize(rb_graph_one->sizeHint());
   graph_bg->insert(rb_graph_sep);
   graph_bg->insert(rb_graph_grp);
   graph_bg->insert(rb_graph_one);
   graph_grid = new QCheckBox("Grid", graph_box); 
   graph_reverse = new QCheckBox("Reverse Colors", graph_box);
   QLabel *graph_numlab = new QLabel("History Length", graph_box);
   graph_numlab->setMinimumSize(graph_numlab->sizeHint());
   graph_num = new QLineEdit(graph_box);
   graph_num->setMinimumSize(QSize(50, le_height));
   graph_num->setMaximumSize(QSize(10000, le_height));
   
   // setup layout 
   QBoxLayout *lay4 = new QBoxLayout(QBoxLayout::LeftToRight);
   lay2->addLayout(lay4, 1);
   lay4->addWidget(l_update, 0);
   lay4->addWidget(le_update, 0);
   lay4->addWidget(l_sec, 0);
   lay4->addWidget(graph_box, 1);
   QBoxLayout *graph_lay0 = new QBoxLayout(graph_box, QBoxLayout::Down, 5);
   graph_lay0->addSpacing(10);
   QBoxLayout *graph_lay1 = new QBoxLayout(QBoxLayout::LeftToRight, 5);
   graph_lay0->addLayout(graph_lay1);
   QBoxLayout *graph_lay2 = new QBoxLayout(QBoxLayout::Down, 5);
   QBoxLayout *graph_lay3 = new QBoxLayout(QBoxLayout::Down, 5);
   QBoxLayout *graph_lay4 = new QBoxLayout(QBoxLayout::LeftToRight, 5);
   graph_lay1->addLayout(graph_lay2);
   graph_lay1->addLayout(graph_lay3);
   graph_lay2->addWidget(rb_graph_lab);
   graph_lay2->addWidget(rb_graph_sep);
   graph_lay2->addWidget(rb_graph_grp);
   graph_lay2->addWidget(rb_graph_one);   
   graph_lay3->addWidget(graph_grid);
   graph_lay3->addWidget(graph_reverse);
   graph_lay3->addLayout(graph_lay4);
   graph_lay4->addWidget(graph_numlab, 0);
   graph_lay4->addWidget(graph_num, 1);
}


void MainWidget::applyFromSetup() {
   // update variables
   if(rb_graph_sep->isChecked())
     tApp->setValue("graph-mode", GM_SEP);
   if(rb_graph_grp->isChecked())
     tApp->setValue("graph-mode", GM_GRP);
   if(rb_graph_one->isChecked())
     tApp->setValue("graph-mode", GM_ONE);
   tApp->setValue("auto-save", b_autosave->isChecked());
   tApp->setValue("graph-grid", graph_grid->isChecked());
   tApp->setValue("graph-reverse", graph_reverse->isChecked());
   tApp->setValue("ruptime", TString(le_ruptime->text()));
   tApp->setValue("groups-file", TString(le_groups->text()));
   tApp->setValue("memory-file", TString(le_memory->text()));
   tApp->setValue("popup-file", TString(le_pop->text()));
   tApp->setValue("arch-file", TString(le_arc->text()));
   int i = atoi(le_update->text());
   if(i<tApp->intLower("update")) {
      TString mes("The update interval may not be lower than " + TString(tApp->intLower("update")) + (tApp->intLower("update")==1?" second!":"seconds!"));
      QMessageBox::message("Error", mes);
      i = tApp->intLower("update");
   }
   if(i<30) {
      QMessageBox::message("Warning", "Update intervals lower than 60 seconds are not useful when using ruptime,\n because the load of every host in the network is broadcastet only once a minute.\nLower intervals also increase system overhead.");
   }
   tApp->setValue("update", i);
   i = atoi(graph_num->text());
   if(i<tApp->intLower("graph-history")) {
      TString mes("The graph history may not be shorter than " + TString(tApp->intLower("graph-history")) + "!");
      QMessageBox::message("Error", mes);      
      i = tApp->intLower("graph-history");
   }
   tApp->setValue("graph-history", i);
   
   // update timer
   timer->changeInterval(tApp->getInt("update")*1000);
   if(dmode==M_GRAPH) dmode=(MODE_T)-1;
   updatedata();
}


void MainWidget::resetSetup() {
   rb_graph_sep->setChecked(GM_SEP==(GRAPHMODE_T)tApp->getInt("graph-mode"));
   rb_graph_grp->setChecked(GM_GRP==(GRAPHMODE_T)tApp->getInt("graph-mode"));
   rb_graph_one->setChecked(GM_ONE==(GRAPHMODE_T)tApp->getInt("graph-mode"));
   b_autosave->setChecked(tApp->getBool("auto-save"));
   graph_grid->setChecked(tApp->getBool("graph-grid"));
   graph_reverse->setChecked(tApp->getBool("graph-reverse"));
   le_ruptime->setText(tApp->getString("ruptime"));
   le_memory->setText(tApp->getString("memory-file"));
   le_pop->setText(tApp->getString("popup-file"));
   le_arc->setText(tApp->getString("arch-file"));
   le_groups->setText(tApp->getString("groups-file"));
   le_update->setText(TString(tApp->getInt("update")));
   graph_num->setText(TString(tApp->getInt("graph-history")));
}


void MainWidget::defaultSetup() {
   rb_graph_sep->setChecked(true);
   rb_graph_grp->setChecked(false);
   rb_graph_one->setChecked(false);
   b_autosave->setChecked(true);
   graph_grid->setChecked(tApp->boolDefault("graph-grid"));
   graph_reverse->setChecked(tApp->boolDefault("graph-reverse"));
   le_ruptime->setText(tApp->stringDefault("ruptime"));
   le_pop->setText(tApp->stringDefault("popup-file"));
   le_arc->setText(tApp->stringDefault("arch-file"));
   le_memory->setText(tApp->stringDefault("memory-file"));
   le_groups->setText(tApp->stringDefault("groups-file"));
   le_update->setText(TString(tApp->intDefault("update")));
   graph_num->setText(TString(tApp->intDefault("graph-history")));
}


void MainWidget::saveItems() {
   // set non-setup items
   tApp->setValue("startup", mode);
   tApp->setValue("sort", sort);   
   TString graph;
   for(int i=0; i < m.num(); i++)
     if(m[i].graph) graph += m[i].name + ",";
   if(graph) graph.truncate(graph.len()-1);
   tApp->setValue("graph-hosts", graph);   
   
   // save 
   TString rcname;
   while(!tApp->save(rcname)) { // retry loop
      TString mes = "Error while updating file '" + rcname + "'! Make sure this file is writable. Try again?";
      // ask user
      if(!QMessageBox::query("Error", mes)) break; // No
   }
}


void MainWidget::applyAndSaveItems() {
   applyFromSetup();
   saveItems();
}


// mainwidget

MainWidget::MainWidget(QWidget *parent, const char *name, WFlags f):
QWidget(parent, name, f) {
   m.fixedSize();
   lay=0;
   glay=0;
   pop=0;
   pop1=0;
   pop2=0;
   pop3=0;
   about_pix=0;
   graph_drag=false;
   no_resize=false;
   createSetup();
   resetSetup();
   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), SLOT(updatedata()));
   connect(this, SIGNAL(mach_changed()), SLOT(updateview()));
   timer->start(tApp->getInt("update")*1000);
   updatedata();
   dmode = (MODE_T)-1; 
   
   // get non setup data
   mode = (MODE_T) tApp->getInt("startup");
   sort = (SORT_T) tApp->getInt("sort");
   TArray<TString> graph = split(tApp->getString("graph-hosts"), ",", false, true);
   for(int i=0; i < graph.num(); i++) {
      graph[i].capitalize();
      for(int j=0; j < m.num(); j++) {
	 if(m[j].name == graph[i]) {
	    m[j].graph = true;
	    break;
	 }
      }
   }
   
   // go
   resort(); // emits mach_changed;
}


MainWidget::~MainWidget() {
   destroy_wid();
   if(tApp->getBool("auto-save"))
     saveItems();
   delete setup;
}


void MainWidget::destroy_wid() {
   if(lay) {
      delete lay;
      lay=0;
   }
   if(glay) {
      delete glay;
      glay=0;
   }
   // crashes:
#if 0
   if(pop) {
      delete pop;
      pop=0;
   }
   if(pop1) {
      delete pop1;
      pop1=0;
   }
   if(pop2) {
      delete pop2;
      pop2=0;
   }
   if(pop3) {
      delete pop3;
      pop3=0;
   }
#endif
   for(int i=0; i < wid.num(); i++) delete wid[i];
   wid.empty();
}


void MainWidget::updateview() {
   // update main
   switch(mode) {
    case M_ABOUT:
      disp_about();
      break;
    case M_LOAD:
      disp_load();
      break;
    case M_GRAPH:
      disp_graph();
      break;
    case M_GROUPS:
      disp_groups();
      break;
   }
   
   // update floating wids
   for(int j=0; j < group_cake_wid.num(); j++) {
      group_cake_wid[j]->clearSlices();
      for(int i=0; i < m.num(); i++) {
	 if(m[i].group == group_cake_grp[j]) {
	    group_cake_wid[j]->setSlice(m[i].name, m[i].load1>=0?m[i].load1:0.0);
	 }
      }         
      group_cake_wid[j]->repaint(0,0,-1,-1,false);
   }
}


void MainWidget::group_pop(const char *name, const QPoint& p) {
   //   QWidget *w = new QWidget(0, 0, WStyle_Customize|WStyle_NoBorder|WStyle_Tool);
   QWidget *w = new QWidget(0);
   QjoCakeDiagram *cake = new QjoCakeDiagram(w);
   QBoxLayout *lay = new QBoxLayout(w, QBoxLayout::Down);
   lay->addWidget(cake);
   cake->setTitle("Group " + TString(name));
   cake->setFormat("<name>\n<value>", "%.1f", "%.2f");
   for(int i=0; i< m.num(); i++) {
      if(m[i].group == name) {
	 cake->setSlice(m[i].name, m[i].load1>=0?m[i].load1:0.0);
      }
   }   
   w->resize(QSize(400,420));
   cake->resize(w->size());
   w->move(p);
   w->show();
   group_cake_wid += cake;
   group_cake_grp += name;
}


void MainWidget::do_about() {
   //  double s = 0.1 + 0.5 * (1.0+sin(double(about_phase*20)*M_PI/180.0)); 
   // printf("***\n");
   QWMatrix mat;
   double s=(about_phase%180)-90;
   if(s<0) s=-s;
   s/=90.0;
   mat.scale(s, s);
   mat.rotate(about_phase);
   
   QPixmap p=about_pix->xForm(mat);
   QPixmap p2(about_lab->width(), about_lab->height());
   p2.fill(black);
   bitBlt(&p2, (about_lab->width()-p.width())/2, 
	  (about_lab->height()-p.height())/2, &p, 0, 0, -1, -1, CopyROP, false);
   bitBlt(about_lab, 0, 0, &p2, 0, 0, -1, -1, CopyROP, false);
   about_phase+=2;
   if(about_phase>=360) about_phase=0;
}


void MainWidget::disp_about() {
   if(about_pix==0) {
      about_pix = new QPixmap(qload_logo);      
   }
   
   if(dmode!=M_ABOUT) {      
      // clean up
      destroy_wid();
      dmode = mode = M_ABOUT;
      erase();
      
      //setup layout
      lay = new QBoxLayout(this, QBoxLayout::Down, 5);
      
      // setup pic
      QWidget *p = new QWidget(this);
      p->setBackgroundColor(black);
      lay->addWidget(p);
      wid += p;      
      about_phase = 0;
      QTimer *t = new QTimer(p);
      connect(t, SIGNAL(timeout()), this, SLOT(do_about())); 
      about_lab = p;
      
      // setup label
      QLabel *l = new QLabel(tApp->getString("application-name") + " V" + 
			     tApp->getString("application-version") + 
			     "\n(C) 1997-1999 by Johannes Overmann", this);
      l->setAlignment(AlignCenter);
      l->setLineWidth(1);
      l->setFrameStyle(QFrame::Panel|QFrame::Raised);
      lay->addWidget(l);
      wid += l;      
      
      // show
      no_resize = true;
      lay->activate();
      no_resize = false;
      l->show();
      p->show();
      t->start(50);
   } 
}


void MainWidget::disp_graph() {
   static TArray<Machine> old_m;
   
   if(dmode==M_GRAPH) { // check for layout change
      if(old_m.num()==m.num()) {
	 int i;
	 for(i=0; i < m.num(); i++) {
	    if(m[i].name != old_m[i].name) break; 
	    if(m[i].group != old_m[i].group) break; 
	    if(m[i].graph != old_m[i].graph) break; 
	 }
	 if(i==m.num()) goto cont; // keep old layout
      }
      dmode=(MODE_T)-1;
   }
   
   cont:
   
   if(dmode!=M_GRAPH) {
      // clean up
      destroy_wid();
      dmode = mode = M_GRAPH;
      erase();
      
      int i,j;
      LoadGraph *p;
      bool grid = tApp->getBool("graph-grid");
      bool rev = tApp->getBool("graph-reverse");
      int num =  tApp->getInt("graph-history");
      
      int numw = 0;
      switch((GRAPHMODE_T)tApp->getInt("graph-mode")) {
       case GM_SEP:
	 for(i=0; i < m.num(); i++) 
	   if(m[i].graph) numw++;
	 break;
       case GM_GRP:
	 for(j=0; j < grp.num(); j++) {
	    int k=0;
	    for(i=0; i < m.num(); i++)
	      if(m[i].graph && (m[i].group==grp[j])) k++;
	    if(k) numw++;
	 }
	 break;
       case GM_ONE:
	 for(i=0; i < m.num(); i++) 
	   if(m[i].graph) numw++;
	 if(numw) numw = 1;
	 break;
      }
      if(numw==0) numw=1;
      int xx;
      if(height()<35) xx = 2;
      else for(xx=1; ((height()-5)/((numw+xx-1)/xx)) < 25+5 ;xx++);
      if(xx > numw) xx = numw;
      int yy = (numw+xx-1)/xx;
      int x=0, y=0;   
      glay = new QGridLayout(this, yy, xx, 5);
      switch((GRAPHMODE_T)tApp->getInt("graph-mode")) {
       case GM_SEP:
	 for(i=0; i < m.num(); i++) {
	    if(m[i].graph) {
	       p = new LoadGraph(this, grid, rev, num, "", &m[i]);
	       glay->addWidget(p, y, x);
	       connect(p, SIGNAL(is_empty()), SIGNAL(mach_changed()));
	       y++;
	       if(y>=yy) {
		  y=0;
		  x++;
	       }
	       wid += p;
	    }
	 }
	 break;
       case GM_GRP:
	 for(j=0; j < grp.num(); j++) {
	    p = new LoadGraph(this, grid, rev, num, grp[j]);
	    for(i=0; i < m.num(); i++)
	      if(m[i].graph && (m[i].group==grp[j])) p->addMachine(&m[i]);
	    if(p->isEmpty()) {
	       delete p;
	    } else {
	       glay->addWidget(p, y, x);
	       connect(p, SIGNAL(is_empty()), SIGNAL(mach_changed()));
	       y++;
	       if(y>=yy) {
		  y=0;
		  x++;
	       }
	       wid += p;
	    }
	 }
	 break;
       case GM_ONE:
	 p = new LoadGraph(this, grid, rev, num, "");
	 for(i=0; i < m.num(); i++)
	   if(m[i].graph) p->addMachine(&m[i]);
	 if(p->isEmpty()) {
	    delete p;
	 } else {
	    glay->addWidget(p, y, x);
	    connect(p, SIGNAL(is_empty()), SIGNAL(mach_changed()));
	    y++;
	    if(y>=yy) {
	       y=0;
	       x++;
	    }
	    wid += p;
	 }
	 break;
      }
      if(wid.num()==0) { // show help
	 QLabel *l = new QLabel("Please select some hosts from the 'Load' screen", this);
	 l->setAlignment(AlignCenter);
	 l->setLineWidth(1);
	 l->setFrameStyle(QFrame::Panel|QFrame::Raised);
	 glay->addMultiCellWidget(l, 0, yy-1, 0, xx-1);
	 wid += l;      
      }
      
      // show
      no_resize = true;
      glay->activate();
      no_resize = false;
      for(i=0; i < wid.num(); i++) wid[i]->show();
      
      // save state
      old_m = m;
   } else { // if(dmode!=M_GRAPH)
      for(int i=0; i < wid.num(); i++) {
	 if(strcmp(wid[i]->className(), "LoadGraph")==0) {
	    ((LoadGraph*)wid[i])->dataChanged();
	    wid[i]->repaint(0,0,-1,-1,false);
	 }
      }
   }
}


void MainWidget::pop1act(int id) {
   switch(mode) {
    case M_GRAPH:
      m[id].graph = true;
      dmode=(MODE_T)-1;
      disp_graph();
      break;
    default:
      break;
   }
}


void MainWidget::pop2act(int) {
}


void MainWidget::pop3act(int) {
}


void MainWidget::toggle_setup() {
   static QRect old;
   static bool oldvalid=false;
   
   if(setup->isVisible()) {
      old = setup->geometry();
      oldvalid = true;
      setup->hide();
   } else {
      if(oldvalid) setup->setGeometry(old);
      setup->show();
   }
}


// machines

int MainWidget::findmach(const TString& name) {
   int i;   
   for(i=0; i < m.num(); i++) {
      if(m[i].name == name) return i;
   }
   m += Machine(name);
   return i;
}


int MainWidget::mustfindmach(const TString& name) {
   int i;   
   for(i=0; i < m.num(); i++) {
      if(m[i].name == name) return i;
   }
   if(i<m.num())
     return i;
   else
     return -1;
}


static int string_cmp(const void *p1, const void *p2) {
   TString *s1 = (TString *)p1;
   TString *s2 = (TString *)p2;
   if(*s1 < *s2) return -1;
   if(*s1 > *s2) return +1;
   return 0;
}


void MainWidget::updatedata() {
   int i,j,k;
   FILE *f;
   TArray<TString> a;
   TArray<TString> l;
   
   // invalidate data
   for(i=0; i<m.num(); i++)
     m[i].invalidateData();
   
   // update data from ruptime
   f = popen(tApp->getString("ruptime"), "r");
   if(f==0) { // error
      timer->stop();
      TString mes("Error while opening pipe to command '" + tApp->getString("ruptime") + "'!\nTerminating auto update.\nEdit Setup and Press Apply to restart.");
      QMessageBox::message("Error", mes);      
   } else {
      a = loadTextFile(f);
      if(a.num()==0) {
	 timer->stop();
	 TString mes("The command '" + tApp->getString("ruptime") + "' prduced no output!\nTerminating auto update.\nEdit Setup and Press Apply to restart.");
	 QMessageBox::message("Error", mes);      
      }
      for(i=0; i<a.num(); i++) {
	 TString line(a[i]);
	 line.translateChar(',', ' ');
	 line.translateChar(':', ' ');
	 line.collapseSpace();
	 if(line.len()==0) continue;
	 
	 // get machine index k
	 for(j=0; isalpha(line[j]); j++);
	 if(j==0) continue; // ignore empty or invalid lines
	 TString name(*line, j);
	 name.capitalize();
	 k = findmach(name); // find and perhaps create machine
	 
	 // parse line
	 if(line.search("down") && line.search("users")==0) {
	    m[k].setLoad(DOWN_LOAD, DOWN_LOAD, DOWN_LOAD);		  
	    continue;
	 }
	 double l1=UNK_LOAD, l2=UNK_LOAD, l3=UNK_LOAD;
	 l = split(line, " ");
	 if(l.num()<4) continue; // invalid line
	 l1 = atof(l[l.num()-3]);
	 l2 = atof(l[l.num()-2]);
	 l3 = atof(l[l.num()-1]);
#if _RANDOM_LOAD_
	 l1 = double(random()%357)/100.0;
	 if(l1<0) l1=0;
#endif
	 // fill in data
	 m[k].setLoad(l1, l2, l3);
      }
      pclose(f);
   }
   
   // update data from ram file
   f = fopen(tApp->getString("memory-file"), "r");
   if(f==0) { // error
      timer->stop();
      TString mes("Error while opening memory file '" + tApp->getString("memory-file") + "'!\nTerminating auto update.\nEdit Setup and Press Apply to restart.");
      QMessageBox::message("Error", mes);      
   } else {
      a = loadTextFile(f);
      for(i=0; i<a.num(); i++) {
	 TString line(a[i]);
	 line.collapseSpace();
	 if(line.len()==0) continue;
	 
	 // get machine index k
	 for(j=0; isalpha(line[j]); j++);
	 if(j==0) continue; // ignore empty or invalid lines
	 TString name(*line, j);
	 name.capitalize();
	 k = findmach(name); // find and perhaps create machine
	 
	 // parse and set data
	 l = split(line, " ");
	 if(l.num()<2) continue; // invalid line
	 j = atoi(l[1]);
	 if(j>0) m[k].ram = j;
      }
      fclose(f);
   }
   
   // update data from group file
   f = fopen(tApp->getString("groups-file"), "r");
   if(f==0) { // error
      timer->stop();
      TString mes("Error while opening groups file '" + tApp->getString("groups-file") + "'!\nTerminating auto update.\nEdit Setup and Press Apply to restart.");
      QMessageBox::message("Error", mes);      
   } else {
      a = loadTextFile(f);
      for(i=0; i<a.num(); i++) {
	 TString line(a[i]);
	 line.collapseSpace();
	 if(line.len()==0) continue;
	 
	 // get machine index k
	 for(j=0; isalpha(line[j]); j++);
	 if(j==0) continue; // ignore empty or invalid lines
	 TString name(*line, j);
	 name.capitalize();
	 k = findmach(name); // find and perhaps create machine
	 
	 // parse and set data
	 l = split(line, " ");
	 if(l.num()<2) continue; // invalid line
	 TString g(l[1]);
	 g.upper();
	 m[k].group = g;	 
      }
      fclose(f);
   }
   
   // update data from arch file
   f = fopen(tApp->getString("arch-file"), "r");
   if(f==0) { // error
      timer->stop();
      TString mes("Error while opening arch file '" + tApp->getString("arch-file") + "'!\nTerminating auto update.\nEdit Setup and Press Apply to restart.");
      QMessageBox::information(0, "Error", mes);      
   } else {
      a = loadTextFile(f);
      for(i=0; i<a.num(); i++) {
	 TString line(a[i]);
	 line.collapseSpace();
	 if(line.len()==0) continue;
	 
	 // get machine index k
	 for(j=0; isalpha(line[j]); j++);
	 if(j==0) continue; // ignore empty or invalid lines
	 TString name(*line, j);
	 name.capitalize();
	 k = findmach(name); // find and perhaps create machine
	 
	 // parse and set data
	 l = split(line, " ");
	 if(l.num()<2) continue; // invalid line
	 TString g(l[1]);
	 //		 g.capitalize();
	 m[k].arch = g;	 
      }
      fclose(f);
   }
   
   // build group history
   grphist.empty();
   grphistl.empty();
   grp.empty();
   for(i=0; i<m.num(); i++) {
      if(m[i].load1>=0.0) {  // valid load ?
	 if(grphist.contains(m[i].group)) { // already there
	    grphist[m[i].group] += 1;
	    grphistl[m[i].group] += m[i].load1;
	 } else { // create new
	    grphist[m[i].group] = 1;
	    grphistl[m[i].group] = m[i].load1;
	    grp += m[i].group;
	 }
      }
   }
   
   // sort group names
   qsort(grp.data(), grp.num(), sizeof(TString), string_cmp);
   
   // update history
   for(i=0; i<m.num(); i++) {
      m[i].hist += m[i].load1;
   }
   
   resort(); // emits mach_changed
}


bool MainWidget::event(QEvent *e) {
   if((e->type()==Event_Resize) && (!no_resize)) {
      if(dmode==M_GRAPH) dmode=(MODE_T)-1;
      updateview();
   }
   return QWidget::event(e);
}



